@@ -8,10 +8,12 @@
 
 #include "config.h"
 
+#include <glib.h>
 #include <gio/gio.h>
 
-#include "xb-node-private.h"
 #include "xb-node-query.h"
+#include "xb-node-silo.h"
+#include "xb-silo-export-private.h"
 #include "xb-silo-query-private.h"
 
 /**
@@ -47,7 +49,8 @@ xb_node_query (XbNode *self, const gchar *xpath, guint limit, GError **error)
  * @query: an #XbQuery
  * @error: the #GError, or %NULL
  *
- * Searches the silo using an prepared query.
+ * Searches the silo using a prepared query. To search using a query with
+ * bound values, use xb_node_query_with_context().
  *
  * It is safe to call this function from a different thread to the one that
  * created the #XbSilo.
@@ -64,7 +67,37 @@ xb_node_query_full (XbNode *self, XbQuery *query, GError **error)
 	g_return_val_if_fail (XB_IS_NODE (self), NULL);
 	g_return_val_if_fail (XB_IS_QUERY (query), NULL);
 	g_return_val_if_fail (error == NULL || *error == NULL, NULL);
-	return xb_silo_query_with_root_full (xb_node_get_silo (self), self, query, error);
+	return xb_silo_query_with_root_full (xb_node_get_silo (self), self, query, NULL, FALSE, error);
+}
+
+/**
+ * xb_node_query_with_context:
+ * @self: a #XbNode
+ * @query: an #XbQuery
+ * @context: (nullable) (transfer none): context including values bound to opcodes of type
+ *     %XB_OPCODE_KIND_BOUND_INTEGER or %XB_OPCODE_KIND_BOUND_TEXT, or %NULL if
+ *     the query doesn’t need any context
+ * @error: the #GError, or %NULL
+ *
+ * Searches the silo using a prepared query, substituting values from the
+ * bindings in @context for bound opcodes as needed.
+ *
+ * It is safe to call this function from a different thread to the one that
+ * created the #XbSilo.
+ *
+ * Please note: Only a subset of XPath is supported.
+ *
+ * Returns: (transfer container) (element-type XbNode): results, or %NULL if unfound
+ *
+ * Since: 0.3.0
+ **/
+GPtrArray *
+xb_node_query_with_context (XbNode *self, XbQuery *query, XbQueryContext *context, GError **error)
+{
+	g_return_val_if_fail (XB_IS_NODE (self), NULL);
+	g_return_val_if_fail (XB_IS_QUERY (query), NULL);
+	g_return_val_if_fail (error == NULL || *error == NULL, NULL);
+	return xb_silo_query_with_root_full (xb_node_get_silo (self), self, query, context, FALSE, error);
 }
 
 /**
@@ -73,7 +106,9 @@ xb_node_query_full (XbNode *self, XbQuery *query, GError **error)
  * @query: an #XbQuery
  * @error: the #GError, or %NULL
  *
- * Searches the silo using an prepared query, returning up to one result.
+ * Searches the silo using a prepared query, returning up to one result. To
+ * search using a query with bound values, use
+ * xb_node_query_first_with_context().
  *
  * It is safe to call this function from a different thread to the one that
  * created the #XbSilo.
@@ -86,6 +121,32 @@ xb_node_query_full (XbNode *self, XbQuery *query, GError **error)
  **/
 XbNode *
 xb_node_query_first_full (XbNode *self, XbQuery *query, GError **error)
+{
+	return xb_node_query_first_with_context (self, query, NULL, error);
+}
+
+/**
+ * xb_node_query_first_with_context:
+ * @self: a #XbNode
+ * @query: an #XbQuery
+ * @context: (nullable) (transfer none): context including values bound to opcodes of type
+ *     %XB_OPCODE_KIND_BOUND_INTEGER or %XB_OPCODE_KIND_BOUND_TEXT, or %NULL if
+ *     the query doesn’t need any context
+ * @error: the #GError, or %NULL
+ *
+ * Searches the silo using a prepared query, returning up to one result.
+ *
+ * It is safe to call this function from a different thread to the one that
+ * created the #XbSilo.
+ *
+ * Please note: Only a subset of XPath is supported.
+ *
+ * Returns: (transfer full): a #XbNode, or %NULL if unfound
+ *
+ * Since: 0.3.0
+ **/
+XbNode *
+xb_node_query_first_with_context (XbNode *self, XbQuery *query, XbQueryContext *context, GError **error)
 {
 	g_autoptr(GPtrArray) results = NULL;
 
@@ -94,7 +155,8 @@ xb_node_query_first_full (XbNode *self, XbQuery *query, GError **error)
 	g_return_val_if_fail (error == NULL || *error == NULL, NULL);
 
 	/* nodes don't have to include themselves as part of the query */
-	results = xb_silo_query_with_root_full (xb_node_get_silo (self), self, query, error);
+	results = xb_silo_query_with_root_full (xb_node_get_silo (self), self, query, context, TRUE, error);
+
 	if (results == NULL)
 		return NULL;
 	return g_object_ref (g_ptr_array_index (results, 0));
@@ -151,16 +213,21 @@ const gchar *
 xb_node_query_text (XbNode *self, const gchar *xpath, GError **error)
 {
 	const gchar *tmp;
-	g_autoptr(XbNode) n = NULL;
+	XbSilo *silo;
+	g_autoptr(GPtrArray) results = NULL;
+	XbSiloNode *sn;
 
 	g_return_val_if_fail (XB_IS_NODE (self), NULL);
 	g_return_val_if_fail (xpath != NULL, NULL);
 	g_return_val_if_fail (error == NULL || *error == NULL, NULL);
 
-	n = xb_node_query_first (self, xpath, error);
-	if (n == NULL)
+	silo = xb_node_get_silo (self);
+	results = xb_silo_query_sn_with_root (silo, self, xpath, 1, error);
+	if (results == NULL)
 		return NULL;
-	tmp = xb_node_get_text (n);
+	sn = g_ptr_array_index (results, 0);
+
+	tmp = xb_silo_get_node_text (silo, sn);
 	if (tmp == NULL) {
 		g_set_error_literal (error,
 				     G_IO_ERROR,
@@ -192,25 +259,30 @@ xb_node_query_text (XbNode *self, const gchar *xpath, GError **error)
 const gchar *
 xb_node_query_attr (XbNode *self, const gchar *xpath, const gchar *name, GError **error)
 {
-	const gchar *tmp;
-	g_autoptr(XbNode) n = NULL;
+	XbSiloNodeAttr *a;
+	XbSilo *silo;
+	g_autoptr(GPtrArray) results = NULL;
+	XbSiloNode *sn;
 
 	g_return_val_if_fail (XB_IS_NODE (self), NULL);
 	g_return_val_if_fail (xpath != NULL, NULL);
 	g_return_val_if_fail (error == NULL || *error == NULL, NULL);
 
-	n = xb_node_query_first (self, xpath, error);
-	if (n == NULL)
+	silo = xb_node_get_silo (self);
+	results = xb_silo_query_sn_with_root (silo, self, xpath, 1, error);
+	if (results == NULL)
 		return NULL;
-	tmp = xb_node_get_attr (n, name);
-	if (tmp == NULL) {
+	sn = g_ptr_array_index (results, 0);
+
+	a = xb_silo_get_node_attr_by_str (silo, sn, name);
+	if (a == NULL) {
 		g_set_error_literal (error,
 				     G_IO_ERROR,
 				     G_IO_ERROR_NOT_FOUND,
 				     "no text data");
 		return NULL;
 	}
-	return tmp;
+	return xb_silo_from_strtab (silo, a->attr_value);
 }
 
 /**
@@ -234,16 +306,25 @@ xb_node_query_attr (XbNode *self, const gchar *xpath, const gchar *name, GError
 gchar *
 xb_node_query_export (XbNode *self, const gchar *xpath, GError **error)
 {
-	g_autoptr(XbNode) n = NULL;
+	GString *xml;
+	XbSilo *silo;
+	g_autoptr(GPtrArray) results = NULL;
+	XbSiloNode *sn;
 
 	g_return_val_if_fail (XB_IS_NODE (self), NULL);
 	g_return_val_if_fail (xpath != NULL, NULL);
 	g_return_val_if_fail (error == NULL || *error == NULL, NULL);
 
-	n = xb_node_query_first (self, xpath, error);
-	if (n == NULL)
+	silo = xb_node_get_silo (self);
+	results = xb_silo_query_sn_with_root (silo, self, xpath, 1, error);
+	if (results == NULL)
 		return NULL;
-	return xb_node_export (n, XB_NODE_EXPORT_FLAG_NONE, error);
+	sn = g_ptr_array_index (results, 0);
+
+	xml = xb_silo_export_with_root (silo, sn, XB_NODE_EXPORT_FLAG_NONE, error);
+	if (xml == NULL)
+		return NULL;
+	return g_string_free (xml, FALSE);
 }
 
 /**
@@ -259,23 +340,26 @@ xb_node_query_export (XbNode *self, const gchar *xpath, GError **error)
  *
  * Please note: Only a subset of XPath is supported.
  *
- * Returns: a guint64, or %G_MAXUINT64 if unfound
+ * Returns: a #guint64, or %G_MAXUINT64 if unfound
  *
  * Since: 0.1.0
  **/
 guint64
 xb_node_query_text_as_uint (XbNode *self, const gchar *xpath, GError **error)
 {
-	g_autoptr(XbNode) n = NULL;
+	const gchar *tmp;
 
 	g_return_val_if_fail (XB_IS_NODE (self), G_MAXUINT64);
 	g_return_val_if_fail (xpath != NULL, G_MAXUINT64);
 	g_return_val_if_fail (error == NULL || *error == NULL, G_MAXUINT64);
 
-	n = xb_node_query_first (self, xpath, error);
-	if (n == NULL)
+	tmp = xb_node_query_text (self, xpath, error);
+	if (tmp == NULL)
 		return G_MAXUINT64;
-	return xb_node_get_text_as_uint (n);
+
+	if (g_str_has_prefix (tmp, "0x"))
+		return g_ascii_strtoull (tmp + 2, NULL, 16);
+	return g_ascii_strtoull (tmp, NULL, 10);
 }
 
 /**
@@ -292,21 +376,24 @@ xb_node_query_text_as_uint (XbNode *self, const gchar *xpath, GError **error)
  *
  * Please note: Only a subset of XPath is supported.
  *
- * Returns: a guint64, or %G_MAXUINT64 if unfound
+ * Returns: a #guint64, or %G_MAXUINT64 if unfound
  *
  * Since: 0.1.0
  **/
 guint64
 xb_node_query_attr_as_uint (XbNode *self, const gchar *xpath, const gchar *name, GError **error)
 {
-	g_autoptr(XbNode) n = NULL;
+	const gchar *tmp;
 
 	g_return_val_if_fail (XB_IS_NODE (self), G_MAXUINT64);
 	g_return_val_if_fail (xpath != NULL, G_MAXUINT64);
 	g_return_val_if_fail (error == NULL || *error == NULL, G_MAXUINT64);
 
-	n = xb_node_query_first (self, xpath, error);
-	if (n == NULL)
+	tmp = xb_node_query_attr (self, xpath, name, error);
+	if (tmp == NULL)
 		return G_MAXUINT64;
-	return xb_node_get_attr_as_uint (n, name);
+
+	if (g_str_has_prefix (tmp, "0x"))
+		return g_ascii_strtoull (tmp + 2, NULL, 16);
+	return g_ascii_strtoull (tmp, NULL, 10);
 }
