@@ -23,11 +23,10 @@ typedef struct {
 	gchar		*element;
 	guint32		 element_idx;
 	GPtrArray	*predicates;	/* of XbStack */
-	gboolean	 is_wildcard;
-	gboolean	 is_parent;
 	XbSiloQueryKind	 kind;
 } XbQuerySection;
 
 GPtrArray	*xb_query_get_sections		(XbQuery	*self);
+gchar		*xb_query_to_string		(XbQuery	*self);
 
 G_END_DECLS
