@@ -12,7 +12,7 @@
 
 #include "xb-node-private.h"
 #include "xb-silo-export-private.h"
-#include "xb-silo-private.h"
+#include "xb-silo-node.h"
 #include "xb-string-private.h"
 
 typedef struct {
@@ -38,58 +38,66 @@ xb_silo_export_node (XbSilo *self, XbSiloExportHelper *helper, XbSiloNode *sn, G
 				xb_silo_from_strtab (self, sn->element_name));
 
 	/* add any attributes */
-	for (guint8 i = 0; i < sn->nr_attrs; i++) {
-		XbSiloAttr *a = xb_silo_get_attr (self, helper->off, i);
+	for (guint8 i = 0; i < xb_silo_node_get_attr_count (sn); i++) {
+		XbSiloNodeAttr *a = xb_silo_node_get_attr (sn, i);
 		g_autofree gchar *key = xb_string_xml_escape (xb_silo_from_strtab (self, a->attr_name));
 		g_autofree gchar *val = xb_string_xml_escape (xb_silo_from_strtab (self, a->attr_value));
 		g_string_append_printf (helper->xml, " %s=\"%s\"", key, val);
 	}
 
-	/* finish the opening tag and add any text if it exists */
-	if (sn->text != XB_SILO_UNSET) {
-		g_autofree gchar *text = xb_string_xml_escape (xb_silo_from_strtab (self, sn->text));
-		g_string_append (helper->xml, ">");
-		g_string_append (helper->xml, text);
+	/* collapse open/close tags together if no text or children */
+	if (helper->flags & XB_NODE_EXPORT_FLAG_COLLAPSE_EMPTY &&
+	    xb_silo_node_get_text_idx (sn) == XB_SILO_UNSET &&
+	    xb_silo_get_child_node (self, sn) == NULL) {
+		g_string_append (helper->xml, " />");
 	} else {
-		g_string_append (helper->xml, ">");
-		if (helper->flags & XB_NODE_EXPORT_FLAG_FORMAT_MULTILINE)
-			g_string_append (helper->xml, "\n");
-	}
-	helper->off += xb_silo_node_get_size (sn);
+		/* finish the opening tag and add any text if it exists */
+		if (xb_silo_node_get_text_idx (sn) != XB_SILO_UNSET) {
+			g_autofree gchar *text = xb_string_xml_escape (xb_silo_get_node_text (self, sn));
+			g_string_append (helper->xml, ">");
+			g_string_append (helper->xml, text);
+		} else {
+			g_string_append (helper->xml, ">");
+			if (helper->flags & XB_NODE_EXPORT_FLAG_FORMAT_MULTILINE)
+				g_string_append (helper->xml, "\n");
+		}
+		helper->off += xb_silo_node_get_size (sn);
 
-	/* recurse deeper */
-	while (xb_silo_get_node(self, helper->off)->is_node) {
-		XbSiloNode *child = xb_silo_get_node (self, helper->off);
-		helper->level++;
-		if (!xb_silo_export_node (self, helper, child, error))
+		/* recurse deeper */
+		while (xb_silo_node_has_flag (xb_silo_get_node (self, helper->off),
+					      XB_SILO_NODE_FLAG_IS_ELEMENT)) {
+			XbSiloNode *child = xb_silo_get_node (self, helper->off);
+			helper->level++;
+			if (!xb_silo_export_node (self, helper, child, error))
+				return FALSE;
+			helper->level--;
+		}
+
+		/* check for the single byte sentinel */
+		sn2 = xb_silo_get_node (self, helper->off);
+		if (xb_silo_node_has_flag (sn2, XB_SILO_NODE_FLAG_IS_ELEMENT)) {
+			g_set_error (error,
+				     G_IO_ERROR,
+				     G_IO_ERROR_INVALID_DATA,
+				     "no seninel at %" G_GUINT32_FORMAT,
+				     helper->off);
 			return FALSE;
-		helper->level--;
-	}
+		}
+		helper->off += xb_silo_node_get_size (sn2);
 
-	/* check for the single byte sentinel */
-	sn2 = xb_silo_get_node (self, helper->off);
-	if (sn2->is_node) {
-		g_set_error (error,
-			     G_IO_ERROR,
-			     G_IO_ERROR_INVALID_DATA,
-			     "no seninel at %" G_GUINT32_FORMAT,
-			     helper->off);
-		return FALSE;
+		/* add closing tag */
+		if ((helper->flags & XB_NODE_EXPORT_FLAG_FORMAT_INDENT) > 0 &&
+		    xb_silo_node_get_text_idx (sn) == XB_SILO_UNSET) {
+			for (guint i = 0; i < helper->level; i++)
+				g_string_append (helper->xml, "  ");
+		}
+		g_string_append_printf (helper->xml, "</%s>",
+					xb_silo_from_strtab (self, sn->element_name));
 	}
-	helper->off += xb_silo_node_get_size (sn2);
-
-	/* add closing tag */
-	if ((helper->flags & XB_NODE_EXPORT_FLAG_FORMAT_INDENT) > 0 &&
-	    sn->text == XB_SILO_UNSET) {
-		for (guint i = 0; i < helper->level; i++)
-			g_string_append (helper->xml, "  ");
-	}
-	g_string_append_printf (helper->xml, "</%s>",
-				xb_silo_from_strtab (self, sn->element_name));
 
 	/* add any optional tail */
-	if (sn->tail != XB_SILO_UNSET) {
-		g_autofree gchar *tail = xb_string_xml_escape (xb_silo_from_strtab (self, sn->tail));
+	if (xb_silo_node_get_tail_idx (sn) != XB_SILO_UNSET) {
+		g_autofree gchar *tail = xb_string_xml_escape (xb_silo_get_node_tail (self, sn));
 		g_string_append (helper->xml, tail);
 	}
 
@@ -101,7 +109,7 @@ xb_silo_export_node (XbSilo *self, XbSiloExportHelper *helper, XbSiloNode *sn, G
 
 /* private */
 GString *
-xb_silo_export_with_root (XbSilo *self, XbNode *root, XbNodeExportFlags flags, GError **error)
+xb_silo_export_with_root (XbSilo *self, XbSiloNode *sroot, XbNodeExportFlags flags, GError **error)
 {
 	XbSiloNode *sn;
 	XbSiloExportHelper helper = {
@@ -117,12 +125,12 @@ xb_silo_export_with_root (XbSilo *self, XbNode *root, XbNodeExportFlags flags, G
 		flags |= XB_NODE_EXPORT_FLAG_INCLUDE_SIBLINGS;
 
 	/* optional subtree export */
-	if (root != NULL) {
-		sn = xb_node_get_sn (root);
+	if (sroot != NULL) {
+		sn = sroot;
 		if (sn != NULL && flags & XB_NODE_EXPORT_FLAG_ONLY_CHILDREN)
-			sn = xb_silo_node_get_child (self, sn);
+			sn = xb_silo_get_child_node (self, sn);
 	} else {
-		sn = xb_silo_get_sroot (self);
+		sn = xb_silo_get_root_node (self);
 	}
 
 	/* no root */
@@ -145,7 +153,7 @@ xb_silo_export_with_root (XbSilo *self, XbNode *root, XbNodeExportFlags flags, G
 		}
 		if ((flags & XB_NODE_EXPORT_FLAG_INCLUDE_SIBLINGS) == 0)
 			break;
-		sn = xb_silo_node_get_next (self, sn);
+		sn = xb_silo_get_next_node (self, sn);
 	} while (sn != NULL);
 
 	/* success */
