@@ -8,11 +8,17 @@
 
 #include <glib-object.h>
 
+#include "xb-query-context.h"
 #include "xb-silo-query.h"
 #include "xb-query.h"
 
 G_BEGIN_DECLS
 
+GPtrArray	*xb_silo_query_sn_with_root	(XbSilo		*self,
+						 XbNode		*n,
+						 const gchar	*xpath,
+						 guint		 limit,
+						 GError		**error);
 GPtrArray	*xb_silo_query_with_root	(XbSilo		*self,
 						 XbNode		*n,
 						 const gchar	*xpath,
@@ -21,6 +27,8 @@ GPtrArray	*xb_silo_query_with_root	(XbSilo		*self,
 GPtrArray	*xb_silo_query_with_root_full	(XbSilo		*self,
 						 XbNode		*n,
 						 XbQuery	*query,
+						 XbQueryContext	*context,
+						 gboolean	 first_result_only,
 						 GError		**error);
 
 G_END_DECLS
