@@ -13,7 +13,8 @@
 
 #include "xb-node-private.h"
 #include "xb-opcode.h"
-#include "xb-silo-private.h"
+#include "xb-opcode-private.h"
+#include "xb-silo-node.h"
 #include "xb-silo-query-private.h"
 #include "xb-stack-private.h"
 #include "xb-query-private.h"
@@ -24,6 +25,9 @@ xb_silo_query_node_matches (XbSilo *self,
 			    XbSiloNode *sn,
 			    XbQuerySection *section,
 			    XbSiloQueryData *query_data,
+			    XbValueBindings *bindings,
+			    guint bindings_offset,
+			    guint *bindings_offset_end_out,
 			    gboolean *result,
 			    GError **error)
 {
@@ -41,20 +45,71 @@ xb_silo_query_node_matches (XbSilo *self,
 	if (section->predicates != NULL) {
 		for (guint i = 0; i < section->predicates->len; i++) {
 			XbStack *opcodes = g_ptr_array_index (section->predicates, i);
-			if (!xb_machine_run (machine, opcodes, result, query_data, error))
+			g_auto(XbValueBindings) predicate_bindings = XB_VALUE_BINDINGS_INIT ();
+			guint predicate_bindings_idx = 0;
+			XbValueBindings *predicate_bindings_ptr = NULL;
+
+			if (bindings != NULL)
+				predicate_bindings_ptr = &predicate_bindings;
+
+			/* set up the bindings for this predicate */
+			for (guint k = 0; bindings != NULL && k < xb_stack_get_size (opcodes); k++) {
+				XbOpcode *op = xb_stack_peek (opcodes, k);
+				if (xb_opcode_is_binding (op)) {
+					/* ignore errors as they’ll be caught by xb_machine_run() */
+					xb_value_bindings_copy_binding (bindings,
+									bindings_offset + predicate_bindings_idx,
+									&predicate_bindings,
+									predicate_bindings_idx);
+					predicate_bindings_idx++;
+				}
+			}
+
+			/* run the predicate; pass NULL for the bindings iff
+			 * (bindings == NULL), as that means we’ve been called
+			 * with pre-0.3.0-style pre-bound values */
+			if (!xb_machine_run_with_bindings (machine, opcodes,
+							   predicate_bindings_ptr,
+							   result,
+							   query_data,
+							   error))
 				return FALSE;
+
+			bindings_offset += predicate_bindings_idx;
 		}
 	}
 
+	if (bindings_offset_end_out != NULL)
+		*bindings_offset_end_out = bindings_offset;
+
 	/* success */
 	return TRUE;
 }
 
+/**
+ * XbSiloQueryHelperFlags:
+ * @XB_SILO_QUERY_HELPER_NONE: No flags set.
+ * @XB_SILO_QUERY_HELPER_USE_SN: Return #XbSiloNodes as results, rather than
+ *    wrapping them in #XbNode. This assumes that they’ll be wrapped later.
+ * @XB_SILO_QUERY_HELPER_FORCE_NODE_CACHE: Always cache the #XbNode objects
+ *
+ * Flags for #XbSiloQueryHelper.
+ *
+ * Since: 0.2.0
+ */
+typedef enum {
+	XB_SILO_QUERY_HELPER_NONE		= 0,
+	XB_SILO_QUERY_HELPER_USE_SN		= 1 << 0,
+	XB_SILO_QUERY_HELPER_FORCE_NODE_CACHE	= 1 << 1,
+} XbSiloQueryHelperFlags;
+
 typedef struct {
 	GPtrArray	*sections;	/* of XbQuerySection */
-	GPtrArray	*results;	/* of XbNode */
+	GPtrArray	*results;	/* of XbNode or XbSiloNode (see @flags) */
+	XbValueBindings	*bindings;
 	GHashTable	*results_hash;	/* of sn:1 */
 	guint		 limit;
+	XbSiloQueryHelperFlags	 flags;
 	XbSiloQueryData	*query_data;
 } XbSiloQueryHelper;
 
@@ -63,7 +118,13 @@ xb_silo_query_section_add_result (XbSilo *self, XbSiloQueryHelper *helper, XbSil
 {
 	if (g_hash_table_lookup (helper->results_hash, sn) != NULL)
 		return FALSE;
-	g_ptr_array_add (helper->results, xb_silo_node_create (self, sn));
+	if (helper->flags & XB_SILO_QUERY_HELPER_USE_SN) {
+		g_ptr_array_add (helper->results, sn);
+	} else {
+		gboolean force_node_cache = (helper->flags & XB_SILO_QUERY_HELPER_FORCE_NODE_CACHE) > 0;
+		g_ptr_array_add (helper->results,
+				 xb_silo_create_node (self, sn, force_node_cache));
+	}
 	g_hash_table_add (helper->results_hash, sn);
 	return helper->results->len == helper->limit;
 }
@@ -75,6 +136,7 @@ static gboolean
 xb_silo_query_section_root (XbSilo *self,
 			    XbSiloNode *sn,
 			    guint i,
+			    guint bindings_offset,
 			    XbSiloQueryHelper *helper,
 			    GError **error)
 {
@@ -92,13 +154,13 @@ xb_silo_query_section_root (XbSilo *self,
 					     "cannot obtain parent for root");
 			return FALSE;
 		}
-		parent = xb_silo_node_get_parent (self, sn);
+		parent = xb_silo_get_parent_node (self, sn);
 		if (parent == NULL) {
 			g_set_error (error,
 				     G_IO_ERROR,
 				     G_IO_ERROR_INVALID_ARGUMENT,
 				     "no parent set for %s",
-				     xb_silo_node_get_element (self, sn));
+				     xb_silo_get_node_element (self, sn));
 			return FALSE;
 		}
 		if (i == helper->sections->len - 1) {
@@ -107,12 +169,12 @@ xb_silo_query_section_root (XbSilo *self,
 		}
 //		g_debug ("PARENT @%u",
 //			 xb_silo_get_offset_for_node (self, parent));
-		return xb_silo_query_section_root (self, parent, i + 1, helper, error);
+		return xb_silo_query_section_root (self, parent, i + 1, bindings_offset, helper, error);
 	}
 
 	/* no node means root */
 	if (sn == NULL) {
-		sn = xb_silo_get_sroot (self);
+		sn = xb_silo_get_root_node (self);
 		if (sn == NULL) {
 			g_set_error_literal (error,
 					     G_IO_ERROR,
@@ -121,7 +183,7 @@ xb_silo_query_section_root (XbSilo *self,
 			return FALSE;
 		}
 	} else {
-		sn = xb_silo_node_get_child (self, sn);
+		sn = xb_silo_get_child_node (self, sn);
 		if (sn == NULL)
 			return TRUE;
 	}
@@ -132,9 +194,12 @@ xb_silo_query_section_root (XbSilo *self,
 	/* continue matching children ".." */
 	do {
 		gboolean result = TRUE;
+		guint bindings_offset_end;
 		query_data->sn = sn;
 		if (!xb_silo_query_node_matches (self, machine, sn, section,
-						 query_data, &result, error))
+						 query_data, helper->bindings,
+						 bindings_offset, &bindings_offset_end,
+						 &result, error))
 			return FALSE;
 		if (result) {
 			if (i == helper->sections->len - 1) {
@@ -144,9 +209,10 @@ xb_silo_query_section_root (XbSilo *self,
 					break;
 			} else {
 //				g_debug ("MATCH %s at @%u, deeper",
-//					 xb_silo_node_get_element (self, sn),
+//					 xb_silo_get_node_element (self, sn),
 //					 xb_silo_get_offset_for_node (self, sn));
-				if (!xb_silo_query_section_root (self, sn, i + 1, helper, error))
+				if (!xb_silo_query_section_root (self, sn, i + 1,
+								 bindings_offset_end, helper, error))
 					return FALSE;
 				if (helper->results->len > 0 &&
 				    helper->results->len == helper->limit)
@@ -166,48 +232,41 @@ xb_silo_query_part (XbSilo *self,
 		    GPtrArray *results,
 		    GHashTable *results_hash,
 		    XbQuery *query,
+		    XbQueryContext *context,
+		    gboolean first_result_only,
 		    XbSiloQueryData *query_data,
+		    XbSiloQueryHelperFlags flags,
 		    GError **error)
 {
+G_GNUC_BEGIN_IGNORE_DEPRECATIONS
 	XbSiloQueryHelper helper = {
 		.results = results,
-		.limit = xb_query_get_limit (query),
+		.bindings = (context != NULL) ? xb_query_context_get_bindings (context) : NULL,
+		.limit = first_result_only ? 1 : (context != NULL ) ? xb_query_context_get_limit (context) : xb_query_get_limit (query),
+		.flags = flags,
 		.results_hash = results_hash,
 		.query_data = query_data,
 	};
+	XbQueryFlags query_flags = (context != NULL) ? xb_query_context_get_flags (context) : xb_query_get_flags (query);
+G_GNUC_END_IGNORE_DEPRECATIONS
 
 	/* find each section */
 	helper.sections = xb_query_get_sections (query);
-	return xb_silo_query_section_root (self, sroot, 0, &helper, error);
+	if (query_flags & XB_QUERY_FLAG_FORCE_NODE_CACHE)
+		helper.flags |= XB_SILO_QUERY_HELPER_FORCE_NODE_CACHE;
+	return xb_silo_query_section_root (self, sroot, 0, 0, &helper, error);
 }
 
-/**
- * xb_silo_query_with_root: (skip)
- * @self: a #XbSilo
- * @n: (allow-none): a #XbNode
- * @xpath: an XPath, e.g. `/components/component[@type=desktop]/id[abe.desktop]`
- * @limit: maximum number of results to return, or 0 for "all"
- * @error: the #GError, or %NULL
- *
- * Searches the silo using an XPath query, returning up to @limit results.
- *
- * It is safe to call this function from a different thread to the one that
- * created the #XbSilo.
- *
- * Please note: Only a subset of XPath is supported.
- *
- * Returns: (transfer container) (element-type XbNode): results, or %NULL if unfound
- *
- * Since: 0.1.0
- **/
-GPtrArray *
-xb_silo_query_with_root (XbSilo *self, XbNode *n, const gchar *xpath, guint limit, GError **error)
+/* Returns an array with (element-type XbSiloNode) if
+ * %XB_SILO_QUERY_HELPER_USE_SN is set, and (element-type XbNode) otherwise. */
+static GPtrArray *
+silo_query_with_root (XbSilo *self, XbNode *n, const gchar *xpath, guint limit, XbSiloQueryHelperFlags flags, GError **error)
 {
 	XbSiloNode *sn = NULL;
 	g_auto(GStrv) split = NULL;
 	g_autoptr(GHashTable) results_hash = g_hash_table_new (g_direct_hash, g_direct_equal);
-	g_autoptr(GPtrArray) results = g_ptr_array_new_with_free_func ((GDestroyNotify) g_object_unref);
-	g_autoptr(GTimer) timer = g_timer_new ();
+	g_autoptr(GPtrArray) results = NULL;
+	g_autoptr(GTimer) timer = xb_silo_start_profile (self);
 	XbSiloQueryData query_data = {
 		.sn = NULL,
 		.position = 0,
@@ -226,6 +285,11 @@ xb_silo_query_with_root (XbSilo *self, XbNode *n, const gchar *xpath, guint limi
 		return NULL;
 	}
 
+	if (flags & XB_SILO_QUERY_HELPER_USE_SN)
+		results = g_ptr_array_new_with_free_func (NULL);
+	else
+		results = g_ptr_array_new_with_free_func ((GDestroyNotify) g_object_unref);
+
 	/* subtree query */
 	if (n != NULL) {
 		sn = xb_node_get_sn (n);
@@ -247,11 +311,15 @@ xb_silo_query_with_root (XbSilo *self, XbNode *n, const gchar *xpath, guint limi
 	for (guint i = 0; split[i] != NULL; i++) {
 		g_autoptr(GError) error_local = NULL;
 		g_autoptr(XbQuery) query = xb_query_new (self, split[i], &error_local);
+		g_auto(XbQueryContext) context = XB_QUERY_CONTEXT_INIT ();
+
 		if (query == NULL) {
 			if (g_error_matches (error_local, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT) &&
 			    (split[i + 1] != NULL || results->len > 0)) {
-				g_debug ("ignoring for OR statement: %s",
-					 error_local->message);
+				if (xb_silo_get_profile_flags (self) & XB_SILO_PROFILE_FLAG_DEBUG) {
+					g_debug ("ignoring for OR statement: %s",
+						 error_local->message);
+				}
 				continue;
 			}
 			g_propagate_prefixed_error (error,
@@ -260,10 +328,12 @@ xb_silo_query_with_root (XbSilo *self, XbNode *n, const gchar *xpath, guint limi
 						    xpath);
 			return NULL;
 		}
-		xb_query_set_limit (query, limit);
+
+		xb_query_context_set_limit (&context, limit);
 		if (!xb_silo_query_part (self, sn,
 					 results, results_hash,
-					 query, &query_data,
+					 query, &context, FALSE, &query_data,
+					 flags,
 					 error)) {
 			return NULL;
 		}
@@ -289,6 +359,53 @@ xb_silo_query_with_root (XbSilo *self, XbNode *n, const gchar *xpath, guint limi
 	return g_steal_pointer (&results);
 }
 
+/**
+ * xb_silo_query_with_root: (skip)
+ * @self: a #XbSilo
+ * @n: (allow-none): a #XbNode
+ * @xpath: an XPath, e.g. `/components/component[@type=desktop]/id[abe.desktop]`
+ * @limit: maximum number of results to return, or 0 for "all"
+ * @error: the #GError, or %NULL
+ *
+ * Searches the silo using an XPath query, returning up to @limit results.
+ *
+ * It is safe to call this function from a different thread to the one that
+ * created the #XbSilo.
+ *
+ * Please note: Only a subset of XPath is supported.
+ *
+ * Returns: (transfer container) (element-type XbNode): results, or %NULL if unfound
+ *
+ * Since: 0.1.0
+ **/
+GPtrArray *
+xb_silo_query_with_root (XbSilo *self, XbNode *n, const gchar *xpath, guint limit, GError **error)
+{
+	return silo_query_with_root (self, n, xpath, limit, XB_SILO_QUERY_HELPER_NONE, error);
+}
+
+/**
+ * xb_silo_query_sn_with_root: (skip)
+ * @self: a #XbSilo
+ * @n: (allow-none): a #XbNode
+ * @xpath: an XPath, e.g. `/components/component[@type=desktop]/id[abe.desktop]`
+ * @limit: maximum number of results to return, or 0 for "all"
+ * @error: the #GError, or %NULL
+ *
+ * A version of xb_silo_query_with_root() which returns results as #XbSiloNodes,
+ * rather than as #XbNodes. This is intended to be used internally to save on
+ * intermediate #XbNode allocations.
+ *
+ * Returns: (transfer container) (element-type XbSiloNode): results, or %NULL if unfound
+ *
+ * Since: 0.2.0
+ **/
+GPtrArray *
+xb_silo_query_sn_with_root (XbSilo *self, XbNode *n, const gchar *xpath, guint limit, GError **error)
+{
+	return silo_query_with_root (self, n, xpath, limit, XB_SILO_QUERY_HELPER_USE_SN, error);
+}
+
 static void
 _g_ptr_array_reverse (GPtrArray *array)
 {
@@ -305,6 +422,11 @@ _g_ptr_array_reverse (GPtrArray *array)
  * @self: a #XbSilo
  * @n: (allow-none): a #XbNode
  * @query: an #XbQuery
+ * @context: (nullable) (transfer none): context including values bound to opcodes of type
+ *     %XB_OPCODE_KIND_BOUND_INTEGER or %XB_OPCODE_KIND_BOUND_TEXT, or %NULL if
+ *     the query doesn’t need any context
+ * @first_result_only: %TRUE if only the first result is going to be used; this
+ *     overrides the limit set in @context, and may perform other optimisations
  * @error: the #GError, or %NULL
  *
  * Searches the silo using an XPath query, returning up to @limit results.
@@ -316,19 +438,22 @@ _g_ptr_array_reverse (GPtrArray *array)
  *
  * Returns: (transfer container) (element-type XbNode): results, or %NULL if unfound
  *
- * Since: 0.1.4
+ * Since: 0.3.0
  **/
 GPtrArray *
-xb_silo_query_with_root_full (XbSilo *self, XbNode *n, XbQuery *query, GError **error)
+xb_silo_query_with_root_full (XbSilo *self, XbNode *n, XbQuery *query, XbQueryContext *context, gboolean first_result_only, GError **error)
 {
 	XbSiloNode *sn = NULL;
 	g_autoptr(GHashTable) results_hash = g_hash_table_new (g_direct_hash, g_direct_equal);
 	g_autoptr(GPtrArray) results = g_ptr_array_new_with_free_func ((GDestroyNotify) g_object_unref);
-	g_autoptr(GTimer) timer = g_timer_new ();
+	g_autoptr(GTimer) timer = xb_silo_start_profile (self);
 	XbSiloQueryData query_data = {
 		.sn = NULL,
 		.position = 0,
 	};
+G_GNUC_BEGIN_IGNORE_DEPRECATIONS
+	XbQueryFlags query_flags = (context != NULL) ? xb_query_context_get_flags (context) : xb_query_get_flags (query);
+G_GNUC_END_IGNORE_DEPRECATIONS
 
 	/* empty silo */
 	if (xb_silo_is_empty (self)) {
@@ -345,17 +470,22 @@ xb_silo_query_with_root_full (XbSilo *self, XbNode *n, XbQuery *query, GError **
 
 	/* only one query allowed */
 	if (!xb_silo_query_part (self, sn, results, results_hash,
-				 query, &query_data, error))
+				 query, context, first_result_only, &query_data,
+				 XB_SILO_QUERY_HELPER_NONE, error))
 		return NULL;
 
 	/* profile */
 	if (xb_silo_get_profile_flags (self) & XB_SILO_PROFILE_FLAG_XPATH) {
 		g_autofree gchar *tmp = xb_query_to_string (query);
+G_GNUC_BEGIN_IGNORE_DEPRECATIONS
+		guint limit = first_result_only ? 1 : (context != NULL) ? xb_query_context_get_limit (context) : xb_query_get_limit (query);
+G_GNUC_END_IGNORE_DEPRECATIONS
+
 		xb_silo_add_profile (self, timer,
 				     "query on %s with `%s` limit=%u -> %u results",
 				     n != NULL ? xb_node_get_element (n) : "/",
 				     tmp,
-				     xb_query_get_limit (query),
+				     limit,
 				     results->len);
 	}
 
@@ -371,7 +501,7 @@ xb_silo_query_with_root_full (XbSilo *self, XbNode *n, XbQuery *query, GError **
 	}
 
 	/* reverse order */
-	if (xb_query_get_flags (query) & XB_QUERY_FLAG_REVERSE)
+	if (query_flags & XB_QUERY_FLAG_REVERSE)
 		_g_ptr_array_reverse (results);
 
 	return g_steal_pointer (&results);
@@ -396,11 +526,37 @@ xb_silo_query_with_root_full (XbSilo *self, XbNode *n, XbQuery *query, GError **
  **/
 GPtrArray *
 xb_silo_query_full (XbSilo *self, XbQuery *query, GError **error)
+{
+	return xb_silo_query_with_context (self, query, NULL, error);
+}
+
+/**
+ * xb_silo_query_with_context:
+ * @self: a #XbSilo
+ * @query: an #XbQuery
+ * @context: (nullable) (transfer none): context including values bound to opcodes of type
+ *     %XB_OPCODE_KIND_BOUND_INTEGER or %XB_OPCODE_KIND_BOUND_TEXT, or %NULL if
+ *     the query doesn’t need any context
+ * @error: the #GError, or %NULL
+ *
+ * Searches the silo using an XPath query.
+ *
+ * It is safe to call this function from a different thread to the one that
+ * created the #XbSilo.
+ *
+ * Please note: Only a subset of XPath is supported.
+ *
+ * Returns: (transfer container) (element-type XbNode): results, or %NULL if unfound
+ *
+ * Since: 0.3.0
+ **/
+GPtrArray *
+xb_silo_query_with_context (XbSilo *self, XbQuery *query, XbQueryContext *context, GError **error)
 {
 	g_return_val_if_fail (XB_IS_SILO (self), NULL);
 	g_return_val_if_fail (XB_IS_QUERY (query), NULL);
 	g_return_val_if_fail (error == NULL || *error == NULL, NULL);
-	return xb_silo_query_with_root_full (self, NULL, query, error);
+	return xb_silo_query_with_root_full (self, NULL, query, context, FALSE, error);
 }
 
 /**
@@ -422,9 +578,37 @@ xb_silo_query_full (XbSilo *self, XbQuery *query, GError **error)
  **/
 XbNode *
 xb_silo_query_first_full (XbSilo *self, XbQuery *query, GError **error)
+{
+	return xb_silo_query_first_with_context (self, query, NULL, error);
+}
+
+/**
+ * xb_silo_query_first_with_context:
+ * @self: a #XbSilo
+ * @query: an #XbQuery
+ * @context: (nullable) (transfer none): context including values bound to opcodes of type
+ *     %XB_OPCODE_KIND_BOUND_INTEGER or %XB_OPCODE_KIND_BOUND_TEXT, or %NULL if
+ *     the query doesn’t need any context
+ * @error: the #GError, or %NULL
+ *
+ * Searches the silo using an XPath query, returning up to one result.
+ *
+ * It is safe to call this function from a different thread to the one that
+ * created the #XbSilo.
+ *
+ * Please note: Only a tiny subset of XPath 1.0 is supported.
+ *
+ * Returns: (transfer none): a #XbNode, or %NULL if unfound
+ *
+ * Since: 0.3.0
+ **/
+XbNode *
+xb_silo_query_first_with_context (XbSilo *self, XbQuery *query, XbQueryContext *context, GError **error)
 {
 	g_autoptr(GPtrArray) results = NULL;
-	results = xb_silo_query_full (self, query, error);
+
+	results = xb_silo_query_with_root_full (self, NULL, query, context, TRUE, error);
+
 	if (results == NULL)
 		return NULL;
 	return g_object_ref (g_ptr_array_index (results, 0));
@@ -511,7 +695,9 @@ xb_silo_query_build_index (XbSilo *self,
 	g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
 	/* do the query */
-	array = xb_silo_query_with_root (self, NULL, xpath, 0, &error_local);
+	array = silo_query_with_root (self, NULL, xpath, 0,
+				      XB_SILO_QUERY_HELPER_USE_SN,
+				      &error_local);
 	if (array == NULL) {
 		if (g_error_matches (error_local,
 				     G_IO_ERROR,
@@ -528,17 +714,16 @@ xb_silo_query_build_index (XbSilo *self,
 
 	/* add each attribute name AND value */
 	for (guint i = 0; i < array->len; i++) {
-		XbNode *n = g_ptr_array_index (array, i);
-		XbSiloNode *sn = xb_node_get_sn (n);
+		XbSiloNode *sn = g_ptr_array_index (array, i);
 		if (attr != NULL) {
-			guint32 off = xb_silo_get_offset_for_node (self, sn);
-			for (guint8 j = 0; j < sn->nr_attrs; j++) {
-				XbSiloAttr *a = xb_silo_get_attr (self, off, j);
+			guint8 attr_count = xb_silo_node_get_attr_count (sn);
+			for (guint8 j = 0; j < attr_count; j++) {
+				XbSiloNodeAttr *a = xb_silo_node_get_attr (sn, j);
 				xb_silo_strtab_index_insert (self, a->attr_name);
 				xb_silo_strtab_index_insert (self, a->attr_value);
 			}
 		} else {
-			xb_silo_strtab_index_insert (self, sn->text);
+			xb_silo_strtab_index_insert (self, xb_silo_node_get_text_idx (sn));
 		}
 	}
 
