@@ -18,6 +18,9 @@ gboolean	 xb_string_contains			(const gchar	*text,
 							 const gchar	*search);
 gboolean	 xb_string_search			(const gchar	*text,
 							 const gchar	*search);
+gboolean	 xb_string_searchv			(const gchar	**text,
+							 const gchar	**search);
+gboolean	 xb_string_token_valid			(const gchar	*text);
 gchar		*xb_string_xml_escape			(const gchar	*str);
 gboolean	 xb_string_isspace			(const gchar	*str,
 							 gssize		 strsz);
